/* tscorel.c */

/*
++++++++++++++++++++++++++++++++++++++++++++++++++
 
.IDENTIFICATION
  program TSCOREL			version 1.00	951208
  A. Kaufer                  		LSW - Heidelberg
  1.00	   951208	

.KEYWORDS
  bulk data frames, time series
 
.PURPOSE
  Xcorrelate a time series of 1-dim spectra
 
.ALGORITHM
  extract input frames from catalogue
  mask and normalize
  rebin to logarithmic scale
  Xcorrelate with template
 
.INPUT/OUTPUT
  the following keys are used:
 
  P1/C/1/64			prefix for outframes
  P3/C/1/64			catalog of frames in time series
  IN_A/C/1/20			template frame 
  IN_B/C/1/20			mask frame
  INPUTI/I/1/1                  1/2 bandwith in integer pixels

    
.VERSIONS
  1.00		first version
--------------------------------------------------
*/

#include <midas_def.h>
#include <math.h>

#define NRANSI
#include "nr.h"
#include "nrutil.h"

#define MAXCAT 1000

int mult_mask (float *tima, int tnpix, double tstart, double tstep,
		float *mima, int mnpix, double mstart, double mstep);
void norm_mean_var (float *ima, int npix);
void rebin_log (float *inima, int npix, double start, double step,
		float *rima, int rnpix, double rstart, double rend, 
		double rstep);
void xcorrelate (float *rima, float *rtima, int rnpix, 
		int nshift, float *xima); 
void statistics(float data[], int n, float *ave, float *adev, float *sdev,
	float *var, float *min, float *max);


main () {

int      frmcnt, iav, iseq, r, i, n;
int      naxis, npix[2], tnaxis, tnpix[2], mnaxis, mnpix[2];
int      xnaxis, xnpix[2], rnaxis, rnpix;
int      imno, timno, mimno, ximno;
int      null, dum, noent, noelem, bytelem, unit, uni, nshift, status; 

char     frame[64], catfil[64], template[64], mask[64], xframe[64], prefix[64];
char     cunit[16*(2+1)], output[64], ident[72];
char     dtype[2], msg[80];

float     *inima, *rima, *tima, *mima, *rtima, *xima;
float     lhcuts[4], min=1.0e34, max=-1.0e34, ave, adev, sdev, var;

double   start[2], step[2], tstart[2], tstep[2], mstart[2], mstep[2];
double   rstart, rend, rstep, xstart[2], xend[2], xstep[2];
double   time[6];
 
/* set up MIDAS environment + enable automatic error abort */

 SCSPRO ("tscorel");

/* get input frame list, template, mask frame  */

 printf ("get names \n");
 SCKGETC ("P1",1L,64L,&iav,prefix);
 SCKGETC ("P3",1L,64L,&iav,catfil);
 SCKGETC ("IN_A",1L,20L,&iav,template);
 SCKGETC ("IN_B",1L,20L,&iav,mask);
 printf ("-> %s %s %s\n",catfil,template,mask);

/*  get the 1/2 bandwith in integer pixels */

 printf ("get bandwidth \n");
 SCKRDI ("INPUTI",1L,1L,&iav,&nshift,&unit,&null);

/*  get template frame  */

 printf ("get template %s\n", template);
 SCIGET (template,D_R4_FORMAT,F_I_MODE,F_IMA_TYPE,1L,&tnaxis,tnpix,
       	 tstart,tstep,ident,cunit,(char **)&tima,&timno);

/*  get mask frame  */

 printf ("get mask %s\n",mask);
 SCIGET (mask,D_R4_FORMAT,F_I_MODE,F_IMA_TYPE,1L,&mnaxis,mnpix,
       	 mstart,mstep,ident,cunit,(char **)&mima,&mimno);

/*  modify template by mask and normalize */

 printf ("modify template by mask and normalize \n");
 status = mult_mask (tima,tnpix[0],tstart[0],tstep[0],
		     mima,mnpix[0],mstart[0],mstep[0]);
 norm_mean_var (tima,tnpix[0]);

/*  take range and step for log rebinning from template 
       this start/step/npix are used for all frames !       */
    
 rstart = log (tstart[0]);
 rend   = log (tstart[0]+(tnpix[0]-1)*tstep[0]);
 rstep  = 1.0*(log(tstart[0]+(tnpix[0]-1)*tstep[0])-
	       log(tstart[0]+(tnpix[0]-2)*tstep[0]));
 rnpix  = (rend-rstart)/(rstep)+1;

 printf ("%6d %12.10f %12.10f \n",rnpix,(float)(rstart),(float)(rstep));

/*  log rebinning of template */

 printf ("rebin template to logarithmic scale \n");
 rtima   = vector (1,rnpix);
 rebin_log (tima,tnpix[0],tstart[0],tstep[0],rtima,rnpix,rstart,rend,rstep);

/*   handle input from catalog  */

 printf ("get noent of catalog \n");
 SCCSHO (catfil,&noent,&dum);
 printf (" -> %d \n",noent);

/* loop on frames in catalogue */

 printf ("loop on catalog \n");
  iseq = 0;
 for (frmcnt = 1; frmcnt <= noent; frmcnt++) {

   SCCGET (catfil,0L,frame,output,&iseq);
   if (frame[0] == ' ') break;		

   SCIGET (frame,D_R4_FORMAT,F_I_MODE,F_IMA_TYPE,1L,&naxis,npix,
	   start,step,ident,cunit,(char **)&inima,&imno);

   /* get time descriptors */

   SCDFND (imno,"DTIME", dtype, &noelem, &bytelem);
   if (strcmp (dtype, " ")!= 0) {
     SCDRDD (imno,"DTIME",1L,6L,&iav,&time[0],&unit,&null);   
   } else {
     sprintf (msg," Warning: descriptor DTIME not present in %s, using O_TIME"
	      , frame);
     SCTPUT (msg);
     SCDRDD (imno,"O_TIME",1L,6L,&iav,&time[0],&unit,&null);
   }

   printf ("%d %s %s %f \n",frmcnt,frame,ident,(float)time[3]);

/*  modify frame by mask and normalize */

   printf ("modify frame by mask and normalize \n");
   status = mult_mask (inima,npix[0],start[0],step[0],
		       mima,mnpix[0],mstart[0],mstep[0]);   
   norm_mean_var (inima,npix[0]);

/*   logarithmic rebinning   */

   printf ("rebin frame to logarithmic scale \n");
   rima   = vector (1,rnpix);
   rebin_log (inima,npix[0],start[0],step[0],rima,rnpix,rstart,rend,rstep);

/* Xcorrelate template & frame, write results to Xcorr image*/

   /* create and open Xcorrelation image             */
   strcpy (xframe,prefix);
   strcat (xframe,frame);
   xnaxis    = 1L;
   xnpix[0]  = 2*nshift+1;

   xstart[0] = 3.e5*(exp(-nshift*rstep)-1);
   xend[0]   = 3.e5*(exp(nshift*rstep)-1);
   xstep[0]  = (xend[0]-xstart[0])/(xnpix[0]-1);
   strcpy (cunit,"    ");

   SCIPUT (xframe,D_R4_FORMAT,F_O_MODE,F_IMA_TYPE,xnaxis,xnpix,
	   xstart,xstep,ident,cunit,(char **)&xima,&ximno);
   SCDWRD (ximno,"DTIME",time,1L,6L,&uni);

   printf ("xcorrelate frame and template \n");
   xcorrelate (rima, rtima, rnpix, nshift, xima); 

   free_vector (rima,1,rnpix);
   SCFCLO (imno); 
   SCFCLO (ximno); 
}

 free_vector (rtima,1,rnpix);
 SCFCLO (timno);
 SCFCLO (mimno);
 SCSEPI ();

} /* end of main */


/*  subroutines   */          

int mult_mask (float *tima, int tnpix, double tstart, double tstep,
	       float *mima, int mnpix, double mstart, double mstep)
{
  int t, t1, t2, toff;
  char msg[80];

   if (fabs(tstep-mstep)<1.e-6) {
     t1 = (int)((mstart-tstart)/tstep)+1;
     t2 = (int)(((mstart+mnpix*mstep)-tstart)/tstep)+1;
     toff = t1;
     /*    printf("%d %d %d\n",t1,t2,tnpix); */

     if (t1>0){
       for (t = 0; t < t1; t++){
	 *(tima+t) = 0.0;
       }
       sprintf (msg," %d pixels set to 0.0 ",t1);
       SCTPUT (msg);
     } else t1 = 0;

     if (tnpix-t2>0){
       for (t = t2; t < tnpix; t++){
	 *(tima+t) = 0.0;
       }
       sprintf (msg," %d pixels set to 0.0 ",tnpix-t2);
       SCTPUT (msg);
     } else t2 = tnpix;

     for (t = t1; t < t2; t++){
       *(tima+t) = (*(tima+t)-1.0)*(*(mima-toff+t));
     }
     return 0;
   } else {
     sprintf (msg," Error: step size of frame and mask do not match");
     SCTPUT (msg);
     return -1;
   }

}


void norm_mean_var (float *ima, int npix)
{
  int i;
  float ave, adev, sdev, var, min, max;

/*  statistics(ima, npix, &ave, &adev, &sdev, &var, &min, &max); */
  statistics ((ima-1), npix, &ave, &adev, &sdev, &var, &min, &max); 
  printf ("Mean = %f, Sdev = %f\n",ave,sdev); 
  for (i = 0; i < npix; i++){ 
    *(ima+i) = (*(ima+i)-ave)/sdev; 
  } 
}

void rebin_log (float *inima, int npix, double start, double step,
		float *rima, int rnpix, double rstart, double rend,
		double rstep)
{
  float *yy, *xint, *xrow;
  float in, out;
  int n, r;

   xrow = vector(1,npix);
   xint = vector(1,npix);

   for (n = 0; n < npix; n++){
    xrow[n] = log (start + n*step);  
    xint[n] = *(inima+n);  
   }

   /* rebin with cubic spline interpolation */

   yy = vector(1,rnpix);
   spline ((xrow-1),(xint-1),npix,0.0,0.0,(yy-1));

   for (r = 0; r < rnpix; r++){
    in = rstart + r*rstep;  
    if (in < xrow[0] || in > xrow[npix-1]) {  /* notwendig ? */
      rima[r] = 0.0;
    } else {
      splint ((xrow-1),(xint-1),(yy-1),npix,in,&out);
      rima[r] = out;
    }
  }

   free_vector (yy,1,rnpix);  
   free_vector (xint,1,npix);  
   free_vector (xrow,1,npix);  
}

void xcorrelate (float *rima, float *rtima, int rnpix,
		 int nshift, float *xima)
{
  int i,n;
  float sum;

  for (n = -nshift; n <= nshift; n++){
    sum = 0.0;
    for (i = 0; i < rnpix; i++){
      sum += *(rima+i)*(*(rtima+i-n));
    }
    *(xima+n+nshift) = sum;
  }
} 

void statistics (float data[], int n, float *ave, float *adev, float *sdev,
		 float *var, float *min, float *max)
{
  void nrerror(char error_text[]);
  int j;
  float ep=0.0,s,p;
  char msg[20];

  if (n <= 1) nrerror("n must be at least 2 in moment");
  s=0.0;
  for (j=1;j<=n;j++){
    if (data[j] > *max) *max = data[j];
    if (data[j] < *min) *min = data[j];
    s += data[j];
  }
  *ave=s/n;
  *adev=(*var)=0.0;
  for (j=1;j<=n;j++) {
    *adev += fabs(s=data[j]-(*ave));
    *var += (p=s*s);
  }
  *adev /= n;
  *var=(*var-ep*ep/n)/(n-1);
  *sdev=(float)sqrt((double)*var);
}


void spline (float x[], float y[], int n, float yp1, float ypn, float y2[])
{
  int i,k;
  float p,qn,sig,un,*u;
  
  u=vector(1,n-1);
  if (yp1 > 0.99e30)
    y2[1]=u[1]=0.0;
  else {
    y2[1] = -0.5;
    u[1]=(3.0/(x[2]-x[1]))*((y[2]-y[1])/(x[2]-x[1])-yp1);
  }
  for (i=2;i<=n-1;i++) {
    sig=(x[i]-x[i-1])/(x[i+1]-x[i-1]);
    p=sig*y2[i-1]+2.0;
    y2[i]=(sig-1.0)/p;
    u[i]=(y[i+1]-y[i])/(x[i+1]-x[i]) - (y[i]-y[i-1])/(x[i]-x[i-1]);
    u[i]=(6.0*u[i]/(x[i+1]-x[i-1])-sig*u[i-1])/p;
  }
  if (ypn > 0.99e30)
    qn=un=0.0;
  else {
    qn=0.5;
    un=(3.0/(x[n]-x[n-1]))*(ypn-(y[n]-y[n-1])/(x[n]-x[n-1]));
  }
  y2[n]=(un-qn*u[n-1])/(qn*y2[n-1]+1.0);
  for (k=n-1;k>=1;k--)
    y2[k]=y2[k]*y2[k+1]+u[k];
  free_vector(u,1,n-1);
}

/* (C) Copr. 1986-92 Numerical Recipes Software *1#R.+,V+%. */



void splint (float xa[], float ya[], float y2a[], int n, float x, float *y)
{
  void nrerror(char error_text[]);
  int klo,khi,k,i;
  float h,b,a;
  
  klo=1;
  khi=n;
  while (khi-klo > 1) {
    k=(khi+klo) >> 1;
    if (xa[k] > x) khi=k;
    else klo=k;
  }
  h=xa[khi]-xa[klo];
  if (h == 0.0) nrerror("Bad xa input to routine splint");
  a=(xa[khi]-x)/h;
  b=(x-xa[klo])/h;
  *y=a*ya[klo]+b*ya[khi]+((a*a*a-a)*y2a[klo]+(b*b*b-b)*y2a[khi])*(h*h)/6.0;
}
/* (C) Copr. 1986-92 Numerical Recipes Software *1#R.+,V+%. */


#include "nrutil.c"

#undef NRANSI




