/* tsminmax.c */

/*
  ++++++++++++++++++++++++++++++++++++++++++++++++++
 
  .IDENTIFICATION
  program TSMINMAX			version 1.00  010917
  O. Stahl                  		LSW - Heidelberg
  1.00	   010917	

  .KEYWORDS
  bulk data frames, time series
 
  .PURPOSE
  find min/max flux spectrum in time series
 
  .ALGORITHM
  extract input frames from catalogue,
  renorm extracted frame 
  filter median 3x1 in t direction 
  find min/max flux spectrum in time series
 
  .INPUT/OUTPUT
  the following keys are used:
 
  OUT_A/C/1/60			result frame
  P3/C/1/80			catalog of frames in time series
  INPUTD/D/1/2                  lower/upper x limit 
  INPUTI/I/1/2                  catalog interval
  P6/C/1/1                      renorm flag
  P7/C/1/1                      filter flag  
    

  .VERSIONS
  1.00		first version
  --------------------------------------------------
*/

#include <midas_def.h>
#include <math.h>

#define NRANSI
#include "nr.h"
#include "nrutil.h"

#define MAXCAT 1000
#define MAXWIN 25
#define HMAXMED 1
#define MAXDIFF 6

void median (float x[],int n,float *xmed);
void iline (float x[], float y[], int n, float a[]);
void ilint (float xa[], float ya[], float a[], int n, float x, float *y);
void minmax(float mm[], float ima[], int npix);

main() {

  int    renorm=0, filter=0;

  int      t_limits[2];  
  int      frmcnt, iav, iseq, x, xstart, t, tt, tmax, r, i, n;
  int      naxis, npix[2], tnaxis, tnpix[2], rnaxis, rnpix[2];
  int      imno, imnot, imnor1, imnor2, null, dum, noent, noelem;
  int      ival[MAXCAT][2], maxintcnt, intcnt, istart, iend, inpix;
  int      unit, uni; 

  char     flag[2], dtype[2];
  char     line[84];
  char     msg[80];
  char     frame[64], *rframe, *rframe1, *rframe2, catfil[64];
  char     cunit[16*(2+1)], ident[72], output[64], history[64];

  float    *inima, *tempima, *resima1, *resima2, *tcol;
  float    yp1, ypn, in, out;
  float    medwin[2*HMAXMED+1], leftwin[MAXWIN], rightwin[MAXWIN];
  float    med, medrwin, medlwin, a, b;
  float    lhcuts[4], min=1.0e34, max=-1.0e34,  var, mm[2];

  double   start[2], step[2], tstart[2], tstep[2], rstart[2], rstep[2];
  double   x_limits[3], t_step;
 
  /* set up MIDAS environment + enable automatic error abort */

  rframe  = (char *) osmmget(64 * sizeof(char));
  rframe1 = (char *) osmmget(64 * sizeof(char));
  rframe2 = (char *) osmmget(64 * sizeof(char));
  
  
  SCSPRO("tsminmax");

  /* get result frame, history, input frame list, flags  */

  SCKGETC("OUT_A",1L,60L,&iav,rframe);
  SCKGETC("HISTORY",1L,64L,&iav,history);
  SCKGETC("P3",1L,80L,&iav,line);
  SCKGETC("P6",1L,1L,&iav,flag);
  if (strcmp(flag,"y")==0 || strcmp(flag,"Y")==0) 
    renorm=1; 
  SCKGETC("P7",1L,1L,&iav,flag);
  if (strcmp(flag,"y")==0 || strcmp(flag,"Y")==0) 
    filter=1; 


  /*  get the data intervals, time_step  */

  SCKRDD("INPUTD",1L,2L,&iav,x_limits,&unit,&null);
  SCKRDI("INPUTI",1L,2L,&iav,t_limits,&unit,&null);

  if (x_limits[1] < x_limits[0]){
    dum = x_limits[1];
    x_limits[1] = x_limits[0];
    x_limits[0] = dum;
    sprintf(msg," correction: using x-interval from %6.1f to %6.1f  \n",x_limits[0],x_limits[1]);
    SCTPUT (msg);
  }

  /*   handle input from catalog - get name of first image file in catalog  */

  strcpy(catfil,line);

  SCCSHO(catfil,&noent,&dum);

  if (t_limits[1] < t_limits[0]){
    dum = t_limits[1];
    t_limits[1] = t_limits[0];
    t_limits[0] = dum;
    sprintf(msg," correction: using catalog entries %d - %d \n",t_limits[0],t_limits[1]);
    SCTPUT (msg);
  }

  if (noent < t_limits[1]) {
    t_limits[1] = noent;
    sprintf(msg," correction: using catalog entries %d - %d \n",t_limits[0],t_limits[1]);
    SCTPUT (msg);
  }

  /* default: use all active entries    */
  if (t_limits[0]==0 || t_limits[1]==0) {
    t_limits[0] = 1;
    t_limits[1] = noent;
  }

iseq = 0;
  for (frmcnt = 1; frmcnt < t_limits[0]; frmcnt++) {
    SCCGET(catfil,0L,frame,output,&iseq);
    if (frame[0] == ' ') break;		
  }


  iseq = 0;
  t = 0; 
  for (frmcnt = t_limits[0]; frmcnt <= t_limits[1]; frmcnt++) {

    SCCGET(catfil,0L,frame,output,&iseq);
    if (frame[0] == ' ') break;		

    SCIGET(frame,D_R4_FORMAT,F_I_MODE,F_IMA_TYPE,1L,&naxis,npix,
	   start,step,ident,cunit,(char **)&inima,&imno);
    
    
    /*    take the step in x from first frame and create temporary frame  
	  with 2 dimensions (x,trow)                                           */
    if (t==0) {
      SCDRDC(imno,"IDENT",1L,1L,72L,&iav,ident,&unit,&null);
      tnaxis = 2L;
      tstart[0] = x_limits[0];
      tstep[0] = step[0];
      tnpix[0] = (long int)(x_limits[1]-x_limits[0]) / tstep [0] + 1;
      tnpix[1] = (long int)(t_limits[1]-t_limits[0]) + 1;
      SCIPUT("middumma",D_R4_FORMAT,F_O_MODE,F_IMA_TYPE,tnaxis,tnpix,
             tstart,tstep,ident,cunit,(char **)&tempima,&imnot);
    }  
  
   
    /*   extract selected wavelength range and insert in temp image           */   

    xstart = (long int)((x_limits[0] - start[0]) / step[0]);
    for (x = 0; x < tnpix[0]; x++) {
      *(tempima+x+t*tnpix[0]) = *(inima+xstart+x);
    }
    t++;       
    SCFCLO(imno); 
  }

  if (renorm) {

    /* find local continuum and renorm temp_ima  */

    for (t=0; t<tnpix[1]; t++){
      for (i = 0; i < MAXWIN; i++){
	leftwin[i]  = *(tempima+i+t*tnpix[0]); 
	rightwin[i] = *(tempima-(i+1)+(t+1)*tnpix[0]);
      }  
      median(leftwin-1,MAXWIN,&medlwin);
      median(rightwin-1,MAXWIN,&medrwin);
      a = (medrwin-medlwin)/(tnpix[0]-0-MAXWIN);
      b = medlwin - a * MAXWIN/2;
      for (x = 0; x < tnpix[0]; x++) {
	*(tempima+x+t*tnpix[0]) = *(tempima+x+t*tnpix[0]) / (a * x + b);      
      }
    }
  }

  if (filter) {

    /* filter median 1x3 in t direction         */

    tcol = vector(1,tnpix[1]);

    for (x = 0; x < tnpix[0]; x++){
      for (t = 0; t < tnpix[1]; t++) {
	*(tcol+t) = *(tempima+t*tnpix[0]+x);
      }

      for (t = HMAXMED; t < tnpix[1]-HMAXMED; t++) {
	for (i = 0; i < 2*HMAXMED+1; i++){
	  medwin[i] = *(tcol+t+i-HMAXMED);
	}
	median(medwin-1,2*HMAXMED+1,&med);
	*(tempima+t*tnpix[0]+x) = med;
      }
    }

    strcpy(rframe1,rframe);
    strcat(rframe1,"_MAX");

    strcpy(rframe2,rframe);
    strcat(rframe2,"_MIN");
    
    SCIPUT (rframe1,D_R4_FORMAT,F_O_MODE,F_IMA_TYPE,1,tnpix,
	    tstart,tstep,ident,cunit,(char **)&resima1,&imnor1);

    SCIPUT (rframe2,D_R4_FORMAT,F_O_MODE,F_IMA_TYPE,1,tnpix,
	    tstart,tstep,ident,cunit,(char **)&resima2,&imnor2);

    for (i=0; i<tnpix[0]; i++)
      {
	resima1[i] = max;
	resima2[i] = min;
      }
    
    for (x=0;x<tnpix[0]; x++){

      for (t = 0; t < tnpix[1]; t++){   
	if (tempima[t*tnpix[0]+x] > resima1[x])
	{
	  resima1[x] = tempima[t*tnpix[0]+x];
	}
	if (tempima[t*tnpix[0]+x] < resima2[x])
	{
	  resima2[x] = tempima[t*tnpix[0]+x];
	}
      }
    }
    
    free_vector(tcol,1,inpix);


    /* compute and write additional descriptors of image header */

    minmax(mm, resima1, tnpix[0]*tnpix[1]);
    lhcuts[3] = mm[0];
    lhcuts[4] = mm[1];
    SCDWRR (imnor1,"LHCUTS",lhcuts,1L,4L,&uni);

    minmax(mm, resima2, tnpix[0]*tnpix[1]);
    lhcuts[3] = mm[0];
    lhcuts[4] = mm[1];
    SCDWRR (imnor2,"LHCUTS",lhcuts,1L,4L,&uni);

    SCSEPI();

  } /* end of main */

}
/*  subroutines   */          

void median (float x[],int n,float *xmed)
{
  int n2,n2p;
  
  sort(n,x);
  n2p=(n2=n/2)+1;
  *xmed=(n % 2 ? x[n2p] : 0.5*(x[n2]+x[n2p]));
}

#define SWAP(a,b) temp=(a);(a)=(b);(b)=temp;
#define M 7
#define NSTACK 50
 
void sort(unsigned long n, float arr[])
{
  unsigned long i,ir=n,j,k,l=1;
  int jstack=0,*istack;
  float a,temp;
  
  istack=ivector(1,NSTACK);
  for (;;) {
    if (ir-l < M) {
      for (j=l+1;j<=ir;j++) {
	a=arr[j];
	for (i=j-1;i>=1;i--) {
	  if (arr[i] <= a) break;
	  arr[i+1]=arr[i];
	}
	arr[i+1]=a;
      }
      if (jstack == 0) break;
      ir=istack[jstack--];
      l=istack[jstack--];
    } else {
      k=(l+ir) >> 1;
      SWAP(arr[k],arr[l+1])
	if (arr[l+1] > arr[ir]) {
	  SWAP(arr[l+1],arr[ir])
	    }
      if (arr[l] > arr[ir]) {
	SWAP(arr[l],arr[ir])
	  }
      if (arr[l+1] > arr[l]) {
	SWAP(arr[l+1],arr[l])
	  }
      i=l+1;
      j=ir;
      a=arr[l];
      for (;;) {
	do i++; while (arr[i] < a);
	do j--; while (arr[j] > a);
	if (j < i) break;
	SWAP(arr[i],arr[j]);
      }
      arr[l]=arr[j];
      arr[j]=a;
      jstack += 2;
      if (jstack > NSTACK) nrerror("NSTACK too small in sort.");
      if (ir-i+1 >= j-l) {
	istack[jstack]=ir;
	istack[jstack-1]=i;
	ir=j-1;
      } else {
	istack[jstack]=j-1;
	istack[jstack-1]=l;
	l=i;
      }
    }
  }
  free_ivector(istack,1,NSTACK);
}

#undef M
#undef NSTACK
#include "nrutil.c"

#undef NRANSI

void minmax(float mm[], float ima[], int npix)
{

  int i;
  
  mm[0] = ima[0];
  mm[1] = ima[0];
  
  for (i=0; i<npix; i++)
    {
    if (ima[i] < mm[0])
      mm[0] = ima[i];
    if (ima[i] > mm[0])
      mm[1] = ima[i];
    }
  
}
